/*
 * Creator: Naman Dixit
 * Notice: © Copyright 2019 Naman Dixit
 */

#include "aux.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>

#if 0
# define log(...) printf("Log: " __VA_ARGS__)
#else
# define log(...)
#endif

#include "xml.h"

Char* tokenGet(Char *line, Char separator, Size *position)
{
    Char *elem = NULL;

    Size i = *position;
    Size line_len = strlen(line);
    while ((i < line_len) && (line[i] != '\n') && (line[i] != separator)) {
        sbufPrint(elem, "%c", line[i]);
        i++;
    }

    sbufPrint(elem, "%c", '\0');
    *position = i + 1;

    return elem;
}

int main(int argc, char** argv)
{
    B32 interactive_mode = false;
    Char *port = NULL;

    for (Size i = 1; i < (Size)argc; i++) {
        Size arg_pos = 0;
        if ((arg_pos = strprefix("-i", argv[i]))) {
            interactive_mode = true;
        } else if ((arg_pos = strprefix("--interactive", argv[i]))) {
            interactive_mode = true;
        } else if ((arg_pos = strprefix("-port=", argv[i]))) {
            port = &(argv[i][arg_pos]);
        }
    }

    if (port == NULL) port = "8080";

    FILE *file_input = NULL, *file_output = NULL;

    if (interactive_mode == false) {
        if (argc != 3)  {
            printf("Help: Incorrect command line parameters\n");
            printf("Help: See README.txt file for info on correct invocation\n");
            exit(-1);
        }

        Char *file_input_name = argv[1];
        Char *file_output_name = argv[2];

        if((file_input = fopen(file_input_name, "r")) == NULL) {
            fprintf(stderr, "Error: The input file %s cannot be opened\n", file_input_name);
            exit(-1);
        }

        if((file_output = fopen(file_output_name, "w")) == NULL) {
            fprintf(stderr, "Error: The output file %s cannot be opened\n", file_output_name);
            exit(-1);
        }
    } else {
        file_input = stdin;
        file_output = stdout;
    }

    Sint sock_fd = socket(AF_INET, SOCK_STREAM, 0);

    struct addrinfo hints = {.ai_family = AF_INET,
                             .ai_socktype = SOCK_STREAM};
    struct addrinfo *result = NULL;
    Sint s = getaddrinfo(NULL, "8080", &hints, &result);
    if (s != 0) {
        fprintf(stderr, "Error: getaddrinfo: %s\n", gai_strerror(s));
        exit(-1);
    }

    while (connect(sock_fd, result->ai_addr, result->ai_addrlen) == -1) {
        fprintf(stderr, "Error: Couldn't connect on port %s, trying again in one second...\n", port);
        sleep(1);
    }

    printf("Log: Starting communication with server on port %s...\n", port);

    while (true) {

        if (interactive_mode == true) {
            printf("===QUERY: ");
        }

        Char *command = NULL;
        { // Read command
            Sint c = 0;
            while (((c = getc(file_input)) != '\n') && (c != EOF))  {
                sbufPrint(command, "%c", (Char)c);
            }

            if (interactive_mode && (c == EOF)) {
                printf("\n");
                break;
            }

            if (command == NULL) {
                break;
            }

            sbufPrint(command, "%c", '\0');
        }

        if (interactive_mode == true) {
            printf("RESPONSE: ");
        }

        Char *xml_request_message = NULL;
        // Response message will be of proper legth, since server will make sure of that.
        Size response_size = KiB(300);
        Char *xml_response_message = calloc(response_size, sizeof(*xml_response_message));

        Size index = 0;
        Char *query_type = tokenGet(command, ',', &index);
        Char *query_key = tokenGet(command, ',', &index);

        if(strcmp(query_type, "GET") == 0) {
            xml_request_message = xmlCreateMessage(XML_Message_Kind_GET,
                                                   query_key, NULL, NULL);

            write(sock_fd, xml_request_message, strlen(xml_request_message));
            read(sock_fd, xml_response_message, response_size);
        } else if(strcmp(query_type, "SET") == 0) {
            Char *query_value = tokenGet(command, '\n', &index);

            xml_request_message = xmlCreateMessage(XML_Message_Kind_PUT,
                                                   query_key, query_value, NULL);
            sbufDelete(query_value);

            write(sock_fd, xml_request_message, strlen(xml_request_message));

            read(sock_fd, xml_response_message, response_size);

        } else if(strcmp(query_type,"DEL") == 0) {
            xml_request_message = xmlCreateMessage(XML_Message_Kind_DELETE,
                                                   query_key, NULL, NULL);
            write(sock_fd, xml_request_message, strlen(xml_request_message));

            read(sock_fd, xml_response_message, response_size);
        }

        sbufDelete(query_type);
        sbufDelete(query_key);

        XML_Message msg = xmlParseMessage(xml_response_message);

        switch (msg.kind) {
            case XML_Message_Kind_RESP_GET: {
                fprintf(file_output, "%s,%s\n", msg.key, msg.value);
            } break;
            case XML_Message_Kind_RESP_PUT_DELETE: {
                fprintf(file_output, "success\n");
            } break;
            case XML_Message_Kind_RESP_ERROR: {
                fprintf(file_output, "error,%s\n", msg.error);
            } break;
            default: {
            } break;
        }

        sbufDelete(xml_request_message);
        free(xml_response_message);
        free(msg.key);
        free(msg.value);
        free(msg.error);
    }
    return 0;
}
