typedef struct Cache_Slot {
    Char *key;
    Char *value;
    B8 valid;
    B8 chance;
} Cache_Slot;

typedef struct Cache_Bucket {
    pthread_mutex_t lock;
    Cache_Slot *slots;
} Cache_Bucket;

typedef struct Cache {
    Size bucket_count;
    Size slot_count;
    Size associativity;
    Cache_Bucket buckets[];
} Cache;

global_variable Cache *global_cache;

internal_function
void cacheInit (Size bucket_count, Size associativity)
{
    Size cache_size = sizeof(*global_cache) + (bucket_count * sizeof(global_cache->buckets[0]));
    global_cache = malloc(cache_size);
    memset(global_cache, 0, cache_size);

    for (Size i = 0; i < bucket_count; i++) {
        global_cache->buckets[i].slots = calloc(associativity, sizeof(global_cache->buckets[i].slots[0]));
        pthread_mutex_init(&(global_cache->buckets[i].lock), NULL);
    }

    global_cache->bucket_count = bucket_count;
    global_cache->associativity = associativity;
    global_cache->slot_count = bucket_count * associativity;
}

internal_function
Size cacheGetBucket (U64 hash)
{
    return hash >> (64 - u64Log2(global_cache->bucket_count));
}

internal_function
void cacheInvalidateSlot (Size bucket, Size slot)
{
    pthread_mutex_lock(&(global_cache->buckets[bucket].lock));

    global_cache->buckets[bucket].slots[slot].valid = false;
    free(global_cache->buckets[bucket].slots[slot].key);
    free(global_cache->buckets[bucket].slots[slot].value);
    global_cache->buckets[bucket].slots[slot].key = NULL;
    global_cache->buckets[bucket].slots[slot].value = NULL;

    pthread_mutex_unlock(&(global_cache->buckets[bucket].lock));
}

internal_function
Cache_Slot cacheGetSlot (Size bucket, Size slot)
{
    Cache_Slot result = {0};

    if (global_cache->buckets[bucket].slots[slot].valid) {
        result = global_cache->buckets[bucket].slots[slot];
    }

    return result;
}

internal_function
Char* cacheGetKV (U64 hash, Char *key, Size *slotptr)
{
    Size bucket = cacheGetBucket(hash);

    for (Size i = 0; i < global_cache->associativity; i++) {
        Cache_Slot cs = cacheGetSlot(bucket, i);
        if (cs.valid && (strcmp(cs.key, key) == 0)) {
            if (slotptr != NULL) {
                *slotptr = i;
            }
            return cs.value;
        }
    }

    return NULL;
}

internal_function
void cacheSetKV (U64 hash, Char *key, Char *value)
{
    Size bucket = cacheGetBucket(hash);

    pthread_mutex_lock(&(global_cache->buckets[bucket].lock));

    Cache_Slot *slots = global_cache->buckets[bucket].slots;

    // If data already present in cache
    for (Size i = 0; i < global_cache->associativity; i++) {
        if(slots[i].valid && (strcmp(slots[i].key, key) == 0)) {
            free(slots[i].value);
            slots[i].value = strdup(value);
            slots[i].chance = true;

            pthread_mutex_unlock(&(global_cache->buckets[bucket].lock));

            return;
        }
    }

    // Else if, there is an empty slot
    for (Size i = 0; i < global_cache->associativity; i++) {
        if (slots[i].valid == false) {
            free(slots[i].key);
            slots[i].key = strdup(key);

            free(slots[i].value);
            slots[i].value = strdup(value);

            slots[i].valid = true;
            slots[i].chance = true;

            pthread_mutex_unlock(&(global_cache->buckets[bucket].lock));

            return;
        }
    }

    // Else, evict using second chance algorithm
    while (true) {
        for (Size i = 0; i < global_cache->associativity; i++) {
            if(slots[i].chance == false) {
                free(slots[i].key);
                slots[i].key = strdup(key);

                free(slots[i].value);
                slots[i].value = strdup(value);

                slots[i].valid = true;
                slots[i].chance = true;

                pthread_mutex_unlock(&(global_cache->buckets[bucket].lock));

                return;
            } else {
                slots[i].chance = false;
            }
        }
    }

    pthread_mutex_unlock(&(global_cache->buckets[bucket].lock));
}

void cacheClear (void)
{
    for (Size i = 0; i < global_cache->bucket_count; i++) {
        for (Size j = 0; j < global_cache->associativity; j++) {
            free(global_cache->buckets[i].slots[j].key);
            free(global_cache->buckets[i].slots[j].value);
            global_cache->buckets[i].slots[j].valid = false;
            global_cache->buckets[i].slots[j].chance = false;
        }
    }
}
